from a_sync.a_sync.function cimport _ASyncFunction
cdef class _ASyncBoundMethod(_ASyncFunction):
    cdef readonly object __weakself__
    cdef readonly bint _is_async_def
    cdef readonly object _cache_handle
    cdef object c_map(self, tuple iterables, object concurrency, str task_name, dict kwargs)
    cdef object __c_self__(self)
cdef bint _is_a_sync_instance(object instance)
cdef void _update_cache_timer(str field_name, object instance, _ASyncBoundMethod bound)



import asyncio
import atexit
import signal
import time

import pytest

from a_sync import ProcessPoolExecutor, ThreadPoolExecutor, PruningThreadPoolExecutor
from a_sync.executor import (
    _EXECUTORS,
    _shutdown_all_executors,
)


class DummyExecutor:
    def __init__(self):
        self.shutdown_called = False

    def shutdown(self, wait=True):
        self.shutdown_called = True


@pytest.mark.asyncio_cooperative
async def test_process_pool_executor_run():
    """Test the :class:`ProcessPoolExecutor` by running and submitting the work function asynchronously.

    This test ensures that the :meth:`~ProcessPoolExecutor.run` method of the
    :class:`~ProcessPoolExecutor` returns a coroutine when executed with a synchronous function.

    See Also:
        - :meth:`~ProcessPoolExecutor.run`
        - :func:`asyncio.iscoroutine`
    """
    executor = ProcessPoolExecutor(1)
    coro = executor.run(time.sleep, 0.1)
    assert asyncio.iscoroutine(coro)
    await coro


@pytest.mark.asyncio_cooperative
async def test_thread_pool_executor_run():
    """Test the :class:`ThreadPoolExecutor` by running and submitting the work function asynchronously.

    This test ensures that the :meth:`~ThreadPoolExecutor.run` method of the
    :class:`~ThreadPoolExecutor` returns a coroutine when executed with a synchronous function.

    See Also:
        - :meth:`~ThreadPoolExecutor.run`
        - :func:`asyncio.iscoroutine`
    """
    executor = ThreadPoolExecutor(1)
    coro = executor.run(time.sleep, 0.1)
    assert asyncio.iscoroutine(coro)
    await coro


@pytest.mark.asyncio_cooperative
async def test_pruning_thread_pool_executor_run():
    """Test the :class:`PruningThreadPoolExecutor` by running and submitting the work function asynchronously.

    This test ensures that the :meth:`~PruningThreadPoolExecutor.run` method of the
    :class:`~PruningThreadPoolExecutor` returns a coroutine when executed with a synchronous function.

    See Also:
        - :meth:`~PruningThreadPoolExecutor.run`
        - :func:`asyncio.iscoroutine`
    """
    executor = PruningThreadPoolExecutor(1)
    coro = executor.run(time.sleep, 0.1)
    assert asyncio.iscoroutine(coro)
    await coro


@pytest.mark.asyncio_cooperative
async def test_process_pool_executor_submit():
    """Test the :class:`ProcessPoolExecutor` by submitting the work function.

    This test ensures that the :meth:`~ProcessPoolExecutor.submit` method of the
    :class:`~ProcessPoolExecutor` returns an :class:`asyncio.Future` when executed with a synchronous function.

    Note:
        The :meth:`~ProcessPoolExecutor.submit` method in this context returns an :class:`asyncio.Future`,
        not a :class:`concurrent.futures.Future`. This is specific to the implementation of the executors
        in the `a_sync` library, which adapts the behavior to integrate with the asyncio event loop.

    See Also:
        - :meth:`~ProcessPoolExecutor.submit`
        - :class:`asyncio.Future`
    """
    executor = ProcessPoolExecutor(1)
    fut = executor.submit(time.sleep, 0.1)
    assert isinstance(fut, asyncio.Future)
    await fut


@pytest.mark.asyncio_cooperative
async def test_thread_pool_executor_submit():
    """Test the :class:`ThreadPoolExecutor` by submitting the work function.

    This test ensures that the :meth:`~ThreadPoolExecutor.submit` method of the
    :class:`~ThreadPoolExecutor` returns an :class:`asyncio.Future` when executed with a synchronous function.

    Note:
        The :meth:`~ThreadPoolExecutor.submit` method in this context returns an :class:`asyncio.Future`,
        not a :class:`concurrent.futures.Future`. This is specific to the implementation of the executors
        in the `a_sync` library, which adapts the behavior to integrate with the asyncio event loop.

    See Also:
        - :meth:`~ThreadPoolExecutor.submit`
        - :class:`asyncio.Future`
    """
    executor = ThreadPoolExecutor(1)
    fut = executor.submit(time.sleep, 0.1)
    assert isinstance(fut, asyncio.Future)
    await fut


@pytest.mark.asyncio_cooperative
async def test_pruning_thread_pool_executor_submit():
    """Test the :class:`PruningThreadPoolExecutor` by submitting the work function.

    This test ensures that the :meth:`~PruningThreadPoolExecutor.submit` method of the
    :class:`~PruningThreadPoolExecutor` returns an :class:`asyncio.Future` when executed with a synchronous function.

    Note:
        The :meth:`~PruningThreadPoolExecutor.submit` method in this context returns an :class:`asyncio.Future`,
        not a :class:`concurrent.futures.Future`. This is specific to the implementation of the executors
        in the `a_sync` library, which adapts the behavior to integrate with the asyncio event loop.

    See Also:
        - :meth:`~PruningThreadPoolExecutor.submit`
        - :class:`asyncio.Future`
    """
    executor = PruningThreadPoolExecutor(1)
    fut = executor.submit(time.sleep, 0.1)
    assert isinstance(fut, asyncio.Future)
    await fut


@pytest.mark.asyncio_cooperative
async def test_process_pool_executor_sync_run():
    """Test the :class:`ProcessPoolExecutor` by running and submitting the work function synchronously.

    This test ensures that the :meth:`~ProcessPoolExecutor.run` method of the
    :class:`~ProcessPoolExecutor` returns a coroutine when executed with a synchronous function.

    See Also:
        - :meth:`~ProcessPoolExecutor.run`
        - :func:`asyncio.iscoroutine`
    """
    executor = ProcessPoolExecutor(0)
    coro = executor.run(time.sleep, 0.1)
    assert asyncio.iscoroutine(coro)
    await coro


@pytest.mark.asyncio_cooperative
async def test_thread_pool_executor_sync_run():
    """Test the :class:`ThreadPoolExecutor` by running and submitting the work function synchronously.

    This test ensures that the :meth:`~ThreadPoolExecutor.run` method of the
    :class:`~ThreadPoolExecutor` returns a coroutine when executed with a synchronous function.

    See Also:
        - :meth:`~ThreadPoolExecutor.run`
        - :func:`asyncio.iscoroutine`
    """
    executor = ThreadPoolExecutor(0)
    coro = executor.run(time.sleep, 0.1)
    assert asyncio.iscoroutine(coro)
    await coro


@pytest.mark.asyncio_cooperative
async def test_pruning_thread_pool_executor_sync_run():
    """Test the :class:`PruningThreadPoolExecutor` by running and submitting the work function synchronously.

    This test ensures that the :meth:`~PruningThreadPoolExecutor.run` method of the
    :class:`~PruningThreadPoolExecutor` returns a coroutine when executed with a synchronous function.

    See Also:
        - :meth:`~PruningThreadPoolExecutor.run`
        - :func:`asyncio.iscoroutine`
    """
    executor = PruningThreadPoolExecutor(0)
    coro = executor.run(time.sleep, 0.1)
    assert asyncio.iscoroutine(coro)
    await coro


@pytest.mark.asyncio_cooperative
async def test_process_pool_executor_sync_submit():
    """Test the :class:`ProcessPoolExecutor` by submitting the work function synchronously.

    This test ensures that the :meth:`~ProcessPoolExecutor.submit` method of the
    :class:`~ProcessPoolExecutor` returns an :class:`asyncio.Future` when executed with a synchronous function.

    Note:
        The :meth:`~ProcessPoolExecutor.submit` method in this context returns an :class:`asyncio.Future`,
        not a :class:`concurrent.futures.Future`. This is specific to the implementation of the executors
        in the `a_sync` library, which adapts the behavior to integrate with the asyncio event loop.

    See Also:
        - :meth:`~ProcessPoolExecutor.submit`
        - :class:`asyncio.Future`
    """
    executor = ProcessPoolExecutor(0)
    fut = executor.submit(time.sleep, 0.1)
    assert isinstance(fut, asyncio.Future)
    await fut


@pytest.mark.asyncio_cooperative
async def test_thread_pool_executor_sync_submit():
    """Test the :class:`ThreadPoolExecutor` by submitting the work function synchronously.

    This test ensures that the :meth:`~ThreadPoolExecutor.submit` method of the
    :class:`~ThreadPoolExecutor` returns an :class:`asyncio.Future` when executed with a synchronous function.

    Note:
        The :meth:`~ThreadPoolExecutor.submit` method in this context returns an :class:`asyncio.Future`,
        not a :class:`concurrent.futures.Future`. This is specific to the implementation of the executors
        in the `a_sync` library, which adapts the behavior to integrate with the asyncio event loop.

    See Also:
        - :meth:`~ThreadPoolExecutor.submit`
        - :class:`asyncio.Future`
    """
    executor = ThreadPoolExecutor(0)
    fut = executor.submit(time.sleep, 0.1)
    assert isinstance(fut, asyncio.Future)
    await fut


@pytest.mark.asyncio_cooperative
async def test_pruning_thread_pool_executor_sync_submit():
    """Test the :class:`PruningThreadPoolExecutor` by submitting the work function synchronously.

    This test ensures that the :meth:`~PruningThreadPoolExecutor.submit` method of the
    :class:`~PruningThreadPoolExecutor` returns an :class:`asyncio.Future` when executed with a synchronous function.

    Note:
        The :meth:`~PruningThreadPoolExecutor.submit` method in this context returns an :class:`asyncio.Future`,
        not a :class:`concurrent.futures.Future`. This is specific to the implementation of the executors
        in the `a_sync` library, which adapts the behavior to integrate with the asyncio event loop.

    See Also:
        - :meth:`~PruningThreadPoolExecutor.submit`
        - :class:`asyncio.Future`
    """
    executor = PruningThreadPoolExecutor(0)
    fut = executor.submit(time.sleep, 0.1)
    assert isinstance(fut, asyncio.Future)
    await fut


def test_executor_registration_and_shutdown(monkeypatch):
    """
    Test that executors are registered and shutdown is called on all of them.
    """
    dummy1 = DummyExecutor()
    dummy2 = DummyExecutor()
    _EXECUTORS.clear()
    from a_sync.executor import register_executor

    register_executor(dummy1)
    register_executor(dummy2)
    assert dummy1 in _EXECUTORS
    assert dummy2 in _EXECUTORS
    _shutdown_all_executors()
    assert dummy1.shutdown_called
    assert dummy2.shutdown_called


def test_signal_and_atexit_registration(monkeypatch):
    """
    Test that signal handlers and atexit are registered and chainable.
    """
    called = {"atexit": False, "sigint": False, "sigterm": False}

    def fake_atexit(func):
        called["atexit"] = True
        return func

    def fake_signal(sig, handler):
        if sig == signal.SIGINT:
            called["sigint"] = True
        elif sig == signal.SIGTERM:
            called["sigterm"] = True
        return handler

    monkeypatch.setattr(atexit, "register", fake_atexit)
    monkeypatch.setattr(signal, "signal", fake_signal)
    # Re-register handlers with monkeypatched functions
    from a_sync.executor import _register_executor_shutdown as reg

    reg()
    assert called["atexit"]
    assert called["sigint"]
    assert called["sigterm"]


def test_multiple_executor_shutdown(monkeypatch):
    """
    Test that all registered executors are shut down, even if one raises.
    """

    class FailingExecutor:
        def __init__(self):
            self.shutdown_called = False

        def shutdown(self, wait=True):
            self.shutdown_called = True
            raise RuntimeError("fail")

    dummy = DummyExecutor()
    failing = FailingExecutor()
    _EXECUTORS.clear()
    from a_sync.executor import register_executor

    register_executor(dummy)
    register_executor(failing)
    # Should not raise
    _shutdown_all_executors()
    assert dummy.shutdown_called
    assert failing.shutdown_called


def test_real_executor_registration():
    """
    Test that real executors are registered and can be shut down.
    """
    _EXECUTORS.clear()
    tpe = ThreadPoolExecutor(max_workers=1)
    ppe = ProcessPoolExecutor(max_workers=1)
    pe = PruningThreadPoolExecutor(max_workers=1)
    assert tpe in _EXECUTORS
    assert ppe in _EXECUTORS
    assert pe in _EXECUTORS
    _shutdown_all_executors()
    # No exceptions should be raised



import pytest
import weakref
from asyncio import CancelledError, create_task, get_event_loop, sleep

from a_sync._smart import (
    SmartFuture,
    SmartTask,
    WeakSet,
    set_smart_task_factory,
    shield,
    smart_task_factory,
)


@pytest.mark.asyncio_cooperative
async def test_smart_future_init_no_args():
    fut = SmartFuture()
    assert fut._loop is get_event_loop()
    assert fut._queue is None
    assert fut._key is None
    assert isinstance(fut._waiters, WeakSet)
    assert len(fut._waiters) == 0


@pytest.mark.asyncio_cooperative
async def test_smart_future_init_loop_arg():
    loop = get_event_loop()
    fut = SmartFuture(loop=loop)
    assert fut._loop is loop
    assert fut._queue is None
    assert fut._key is None
    assert isinstance(fut._waiters, WeakSet)
    assert len(fut._waiters) == 0


@pytest.mark.asyncio_cooperative
async def test_smart_future_await():
    fut = SmartFuture()
    get_event_loop().call_soon(fut.set_result, None)
    await fut


@pytest.mark.asyncio_cooperative
async def test_smart_future_await_exc():
    fut = SmartFuture()
    get_event_loop().call_soon(fut.set_exception, ValueError("test"))
    with pytest.raises(ValueError, match="test"):
        await fut


@pytest.mark.asyncio_cooperative
async def test_smart_future_await_cancelled():
    fut = SmartFuture()
    get_event_loop().call_soon(fut.cancel)
    with pytest.raises(CancelledError):
        await fut


@pytest.mark.asyncio_cooperative
async def test_smart_task_await():
    await SmartTask(sleep(0.1), loop=None)


@pytest.mark.asyncio_cooperative
async def test_smart_task_await_exc():
    async def raise_exc():
        raise ValueError("test")

    with pytest.raises(ValueError, match="test"):
        await SmartTask(raise_exc(), loop=None)


@pytest.mark.asyncio_cooperative
async def test_smart_task_await_cancelled():
    task = SmartTask(sleep(0.1), loop=None)
    get_event_loop().call_soon(task.cancel)
    with pytest.raises(CancelledError):
        await task


@pytest.mark.asyncio_cooperative
async def test_smart_task_name():
    t = SmartTask(sleep(0.1), loop=None, name="test")
    assert t.get_name() == "test"


async def smart_task_coro():
    task = create_task(sleep(0.1))
    assert isinstance(task, SmartTask)
    assert await task is None


def test_set_smart_task_factory():
    set_smart_task_factory()
    loop = get_event_loop()
    assert loop.get_task_factory() is smart_task_factory
    loop.run_until_complete(smart_task_coro())


def test_set_smart_task_factory_with_loop():
    loop = get_event_loop()
    set_smart_task_factory(loop)
    assert loop.get_task_factory() is smart_task_factory
    loop.run_until_complete(smart_task_coro())


@pytest.mark.asyncio_cooperative
async def test_shield():
    task = create_task(sleep(1))
    shielded = shield(task)
    await shielded


@pytest.mark.asyncio_cooperative
async def test_shield_exc():
    async def raise_exc():
        raise ValueError("test")

    task = create_task(raise_exc())
    shielded = shield(task)
    with pytest.raises(ValueError, match="test"):
        await shielded


@pytest.mark.asyncio_cooperative
async def test_shield_cancel_inner():
    task = create_task(sleep(1))
    shielded = shield(task)
    task.cancel()
    with pytest.raises(CancelledError):
        await task
    with pytest.raises(CancelledError):
        await shielded


@pytest.mark.asyncio_cooperative
async def test_shield_cancel_outer():
    task = create_task(sleep(1))
    shielded = shield(task)
    shielded.cancel()
    await sleep(0)
    assert not task.cancelled()
    await task
    with pytest.raises(CancelledError):
        await shielded



import asyncio
import pytest

from a_sync.utils.iterators import as_yielded


# Helpers
class AsyncGen:
    def __init__(self, values, delay=0):
        self.values = list(values)
        self.delay = delay

    def __aiter__(self):
        self._idx = 0
        return self

    async def __anext__(self):
        if self._idx >= len(self.values):
            raise StopAsyncIteration
        val = self.values[self._idx]
        self._idx += 1
        if self.delay:
            await asyncio.sleep(self.delay)
        return val


class ErrorGen:
    def __aiter__(self):
        return self

    async def __anext__(self):
        raise Exception("Iterator Error!")


@pytest.mark.asyncio_cooperative
async def test_simple_merge():
    gen1 = AsyncGen([1, 2, 3])
    gen2 = AsyncGen([10, 20])
    results = set()
    async for item in as_yielded(gen1, gen2):
        results.add(item)
    assert results == {1, 2, 3, 10, 20}


@pytest.mark.asyncio_cooperative
async def test_merge_with_empty():
    gen_empty = AsyncGen([])
    gen_full = AsyncGen([5])
    results = []
    async for item in as_yielded(gen_empty, gen_full):
        results.append(item)
    assert results == [5]


@pytest.mark.asyncio_cooperative
async def test_merge_slow_fast():
    fast = AsyncGen([1, 2, 3], delay=0.01)
    slow = AsyncGen(["a", "b"], delay=0.05)
    got = []
    async for item in as_yielded(fast, slow):
        got.append(item)
    # All should appear, timing for order is implementation dependent.
    assert set(got) == {1, 2, 3, "a", "b"}


@pytest.mark.asyncio_cooperative
async def test_error_in_iterator():
    error_gen = ErrorGen()
    ok_gen = AsyncGen([4, 5])
    got = []
    with pytest.raises(Exception, match="Iterator Error!"):
        async for item in as_yielded(error_gen, ok_gen):
            got.append(item)


@pytest.mark.asyncio_cooperative
async def test_single_iterator():
    gen = AsyncGen(range(5))
    results = []
    async for item in as_yielded(gen):
        results.append(item)
    assert results == [0, 1, 2, 3, 4]


@pytest.mark.asyncio_cooperative
async def test_all_empty_iterators():
    gens = [AsyncGen([]) for _ in range(3)]
    results = []
    async for item in as_yielded(*gens):
        results.append(item)
    assert results == []


@pytest.mark.asyncio_cooperative
async def test_consumer_cancels_early():
    # If the consumer stops early, as_yielded should not misbehave or deadlock.
    gen = AsyncGen(range(100))
    count = 0
    async for item in as_yielded(gen):
        count += 1
        if count >= 5:
            break
    assert count == 5



import pytest

from a_sync import property
from a_sync.a_sync.function import _ASyncFunction, ASyncFunction


class PropertyTester:
    def __init__(self, i: int):
        self.i = i

    @property
    async def do_stuff(self):
        return self.i


testers = [PropertyTester(i) for i in range(5)]


def test_property_descriptor_any_sync():
    assert isinstance(PropertyTester.do_stuff.any, _ASyncFunction)
    assert isinstance(PropertyTester.do_stuff.any, ASyncFunction)
    assert PropertyTester.do_stuff.any.is_async_def()
    assert not PropertyTester.do_stuff.any.is_sync_default()
    assert PropertyTester.do_stuff.any(testers, sync=True) == True


@pytest.mark.asyncio_cooperative
async def test_property_descriptor_any_async():
    assert isinstance(PropertyTester.do_stuff.any, _ASyncFunction)
    assert isinstance(PropertyTester.do_stuff.any, ASyncFunction)
    assert PropertyTester.do_stuff.any.is_async_def()
    assert not PropertyTester.do_stuff.any.is_sync_default()
    assert await PropertyTester.do_stuff.any(testers, sync=False) == True


def test_property_descriptor_all_sync():
    assert isinstance(PropertyTester.do_stuff.all, _ASyncFunction)
    assert isinstance(PropertyTester.do_stuff.all, ASyncFunction)
    assert PropertyTester.do_stuff.all.is_async_def()
    assert not PropertyTester.do_stuff.all.is_sync_default()
    assert PropertyTester.do_stuff.all(testers, sync=True) == False


@pytest.mark.asyncio_cooperative
async def test_property_descriptor_all_async():
    assert isinstance(PropertyTester.do_stuff.all, _ASyncFunction)
    assert isinstance(PropertyTester.do_stuff.all, ASyncFunction)
    assert PropertyTester.do_stuff.all.is_async_def()
    assert not PropertyTester.do_stuff.all.is_sync_default()
    assert await PropertyTester.do_stuff.all(testers, sync=False) == False


def test_property_descriptor_min_sync():
    assert isinstance(PropertyTester.do_stuff.min, _ASyncFunction)
    assert isinstance(PropertyTester.do_stuff.min, ASyncFunction)
    assert PropertyTester.do_stuff.min.is_async_def()
    assert not PropertyTester.do_stuff.min.is_sync_default()
    assert PropertyTester.do_stuff.min(testers, sync=True) == 0


@pytest.mark.asyncio_cooperative
async def test_property_descriptor_min_async():
    assert isinstance(PropertyTester.do_stuff.min, _ASyncFunction)
    assert isinstance(PropertyTester.do_stuff.min, ASyncFunction)
    assert PropertyTester.do_stuff.min.is_async_def()
    assert not PropertyTester.do_stuff.min.is_sync_default()
    assert await PropertyTester.do_stuff.min(testers, sync=False) == 0


def test_property_descriptor_max_sync():
    assert isinstance(PropertyTester.do_stuff.max, _ASyncFunction)
    assert isinstance(PropertyTester.do_stuff.max, ASyncFunction)
    assert PropertyTester.do_stuff.max.is_async_def()
    assert not PropertyTester.do_stuff.max.is_sync_default()
    assert PropertyTester.do_stuff.max(testers, sync=True) == 4


@pytest.mark.asyncio_cooperative
async def test_property_descriptor_max_async():
    assert isinstance(PropertyTester.do_stuff.max, _ASyncFunction)
    assert isinstance(PropertyTester.do_stuff.max, ASyncFunction)
    assert PropertyTester.do_stuff.max.is_async_def()
    assert not PropertyTester.do_stuff.max.is_sync_default()
    assert await PropertyTester.do_stuff.max(testers, sync=False) == 4


def test_property_descriptor_sum_sync():
    assert isinstance(PropertyTester.do_stuff.sum, _ASyncFunction)
    assert isinstance(PropertyTester.do_stuff.sum, ASyncFunction)
    assert PropertyTester.do_stuff.sum.is_async_def()
    assert not PropertyTester.do_stuff.sum.is_sync_default()
    assert PropertyTester.do_stuff.sum(testers, sync=True) == 10


@pytest.mark.asyncio_cooperative
async def test_property_descriptor_sum_async():
    assert isinstance(PropertyTester.do_stuff.sum, _ASyncFunction)
    assert isinstance(PropertyTester.do_stuff.sum, ASyncFunction)
    assert PropertyTester.do_stuff.sum.is_async_def()
    assert not PropertyTester.do_stuff.sum.is_sync_default()
    assert await PropertyTester.do_stuff.sum(testers, sync=False) == 10
